#include <cmath>
#include <msp/geometry/box.h>
#include <msp/geometry/circle.h>
#include <msp/geometry/extrudedshape.h>
#include <msp/geometry/halfspace.h>
#include <msp/geometry/intersection.h>
#include <msp/geometry/negation.h>
#include <msp/geometry/transformedshape.h>
#include "trackpart.h"

using namespace std;
using namespace Msp;

namespace R2C2 {

TrackPart::TrackPart():
	length(0),
	radius(0),
	path(0),
	dead_end(false),
	shape(0)
{
	links[0] = 0;
	links[1] = 0;
}

TrackPart::~TrackPart()
{
	delete shape;
}

float TrackPart::get_length() const
{
	if(radius)
		return abs(radius)*length;
	else
		return length;
}

OrientedPoint TrackPart::get_point(float d) const
{
	OrientedPoint result;

	if(radius)
	{
		Angle a = Angle::from_radians(d/radius);
		result.position = transform.transform(Vector(sin(a)*radius, (1-cos(a))*radius, 0));
		result.rotation = dir+a;
	}
	else
	{
		result.position = transform.transform(Vector(d, 0, 0));
		result.rotation = dir;
	}

	return result;
}

OrientedPoint TrackPart::get_nearest_point(const Vector &p) const
{
	OrientedPoint result;
	if(radius)
	{
		Vector v = p-transform.transform(Vector(0, radius, 0));
		Angle a = Geometry::atan2(v.y, v.x)+Angle::quarter_turn()-dir;
		if(radius<0)
			a = Angle::half_turn()-a;
		a.wrap_with_base(Angle::from_radians(length/2)-Angle::half_turn());
		a = min(max(a, Angle::zero()), Angle::from_radians(length));
		if(radius<0)	
			a = -a;
		result.position = transform.transform(Vector(sin(a)*radius, (1-cos(a))*radius, 0));
		result.rotation = dir+a;
	}
	else
	{
		Vector v = p-pos;
		Vector dir_vec = transform.transform_linear(Vector(1, 0, 0));
		float d = min(max(dot(dir_vec, v), 0.0f), length);
		result.position = pos+dir_vec*d;
		result.rotation = dir;
	}
	return result;
}

void TrackPart::check_link(TrackPart &other)
{
	unsigned n_eps = (dead_end ? 1 : 2);
	unsigned n_other_eps = (other.is_dead_end() ? 1 : 2);
	for(unsigned i=0; i<n_eps; ++i)
	{
		OrientedPoint p1 = get_point(i ? get_length() : 0);
		for(unsigned j=0; j<n_other_eps; ++j)
		{
			OrientedPoint p2 = other.get_point(j ? other.get_length() : 0);

			Vector span = p2.position-p1.position;

			Angle da = wrap_balanced(p2.rotation-p1.rotation+Angle::half_turn()*float((i+j+1)%2));

			if(dot(span, span)<1e-6 && abs(da).radians()<=0.01)
			{
				links[i] = &other;
				other.links[j] = this;
				return;
			}
		}
	}
}

TrackPart *TrackPart::get_link(unsigned i) const
{
	if(i>=2)
		throw out_of_range("TrackPart::get_link");
	return links[i];
}

void TrackPart::create_shape()
{
	if(radius)
	{
		Geometry::ExtrudedShape<float, 3> ring(
			Geometry::Intersection<float, 2>(
				Geometry::Circle<float>(abs(radius)+0.02),
				Geometry::Negation<float, 2>(Geometry::Circle<float>(abs(radius)-0.02))),
			0.01);
		Geometry::HalfSpace<float, 3> start_wall(Vector(-1, 0, 0));
		Geometry::HalfSpace<float, 3> end_wall(rotated_vector(Vector(1, 0, 0),
			Angle::from_radians(length*(radius<0 ? -1 : 1))));
		/* This box only exists so the bounding boxes of curves make sense.  It
		can be removed when libmspmath produces better bounding boxes. */
		float w = sin(length)*(abs(radius)+0.02);
		float h = (1-cos(length))*abs(radius-0.02)+0.04;
		Geometry::TransformedShape<float, 3> bounds(
			Geometry::Box<float>(w, h, 0.01),
			Transform::translation(Vector(w/2, (h/2-0.02)*(radius<0 ? -1 : 1)-radius, 0.005)));
		vector<const Shape *> shapes;
		shapes.push_back(&ring);
		shapes.push_back(&start_wall);
		shapes.push_back(&end_wall);
		shapes.push_back(&bounds);
		shape = new Geometry::TransformedShape<float, 3>(
			Geometry::Intersection<float, 3>::from_iterator_range(shapes.begin(), shapes.end()),
			transform*Transform::translation(Vector(0, radius, 0.005)));
	}
	else
	{
		// TODO Get the track profile dimensions from somewhere
		shape = new Geometry::TransformedShape<float, 3>(
			Geometry::Box<float>(length, 0.04, 0.01),
			transform*Transform::translation(Vector(length/2, 0, 0.005)));
	}
}


TrackPart::Loader::Loader(TrackPart &p):
	Msp::DataFile::ObjectLoader<TrackPart>(p)
{
	add("start",    &Loader::start);
	add("length",   &TrackPart::length);
	add("radius",   &TrackPart::radius);
	add("path",     &TrackPart::path);
	add("dead_end", &TrackPart::dead_end);
}

void TrackPart::Loader::finish()
{
	if(obj.radius)
	{
		obj.length *= M_PI/180;
		obj.radius /= 1000;
	}
	else
		obj.length /= 1000;

	obj.pos /= 1000;

	obj.transform = Transform::translation(obj.pos)*
		Transform::rotation(obj.dir, Vector(0, 0, 1));
}

void TrackPart::Loader::start(float x, float y, float d)
{
	obj.pos = Vector(x, y, 0);
	obj.dir = Angle::from_degrees(d);
}

} // namespace R2C2
