#include <cmath>
#include "catalogue.h"
#include "driver.h"
#include "layout.h"
#include "track.h"
#include "trackappearance.h"
#include "trackiter.h"
#include "tracktype.h"
#include "vehicle.h"
#include "vehicletype.h"

using namespace std;
using namespace Msp;

namespace R2C2 {

Vehicle::Vehicle(Layout &l, const VehicleType &t):
	Object(l),
	type(t),
	train(0),
	next(0),
	prev(0),
	placement(type),
	front_sensor(0),
	back_sensor(0)
{
	axles.assign(type.get_axles().begin(), type.get_axles().end());
	for(vector<Axle>::iterator i=axles.begin(); i!=axles.end(); ++i)
		if(!i->type->bogie)
			fixed_axles.push_back(&*i);
	bogies.assign(type.get_bogies().begin(), type.get_bogies().end());
	rods.assign(type.get_rods().begin(), type.get_rods().end());
	for(vector<Bogie>::iterator i=bogies.begin(); i!=bogies.end(); ++i)
		for(unsigned j=0; j<i->axles.size(); ++j)
			i->axles[j] = &axles[i->type->first_axle+j];

	update_rods();

	layout.add(*this);
}

Vehicle::~Vehicle()
{
	if(next)
		detach_back();
	if(prev)
		detach_front();
	layout.remove(*this);
}

Vehicle *Vehicle::clone(Layout *to_layout) const
{
	Vehicle *veh = new Vehicle((to_layout ? *to_layout : layout), type);
	veh->set_position(position);
	veh->set_rotation(rotation);
	return veh;
}

void Vehicle::set_train(Train *t)
{
	train = t;
}

void Vehicle::attach_back(Vehicle &veh)
{
	if(next || veh.prev)
		throw attachment_error("already attached");

	next = &veh;
	veh.prev = this;

	if(is_placed())
		propagate_backward();
}

void Vehicle::attach_front(Vehicle &veh)
{
	if(prev || veh.next)
		throw attachment_error("already attached");

	prev = &veh;
	veh.next = this;

	if(prev->is_placed())
		prev->propagate_backward();
}

void Vehicle::detach_back()
{
	if(!next)
		throw attachment_error("not attached");

	next->prev = 0;
	next = 0;
}

void Vehicle::detach_front()
{
	if(!prev)
		throw attachment_error("not attached");

	prev->next = 0;
	prev = 0;
}

void Vehicle::place(const TrackOffsetIter &t, VehiclePlacement::Anchor a)
{
	if(!t)
		throw invalid_argument("Vehicle::place");
	float gauge_ratio = t->get_type().get_gauge()/type.get_gauge();
	if(gauge_ratio<0.99 || gauge_ratio>1.01)
		throw logic_error("Incompatible gauge");

	placement.place(t, a);

	update_position(0);
	propagate_position();
}

void Vehicle::unplace()
{
	if(!placement.is_placed())
		return;

	placement.unplace();

	if(prev)
		prev->unplace();
	if(next)
		next->unplace();
}

void Vehicle::advance(float d)
{
	placement.advance(d);
	turn_axles(d);
	update_position(d<0 ? -1 : 1);
	propagate_position();
}

const Vehicle::Axle &Vehicle::get_axle(unsigned i) const
{
	if(i>=axles.size())
		throw out_of_range("Vehicle::get_axle");
	return axles[i];
}

const Vehicle::Axle &Vehicle::get_fixed_axle(unsigned i) const
{
	if(i>=fixed_axles.size())
		throw out_of_range("Vehicle::get_fixed_axle");
	return *fixed_axles[i];
}

const Vehicle::Bogie &Vehicle::get_bogie(unsigned i) const
{
	if(i>=bogies.size())
		throw out_of_range("Vehicle::get_bogie");
	return bogies[i];
}

const Vehicle::Axle &Vehicle::get_bogie_axle(unsigned i, unsigned j) const
{
	if(i>=bogies.size())
		throw out_of_range("Vehicle::get_bogie_axle");
	if(j>=bogies[i].axles.size())
		throw out_of_range("Vehicle::get_bogie_axle");
	return *bogies[i].axles[j];
}

const Vehicle::Rod &Vehicle::get_rod(unsigned i) const
{
	if(i>=rods.size())
		throw out_of_range("Vehicle::get_rod");
	return rods[i];
}

void Vehicle::update_position(int sign)
{
	OrientedPoint p = placement.get_point();
	position = p.position;
	// TODO Move the z adjustment to VehiclePlacement
	position.z += placement.get_position(VehiclePlacement::FRONT_AXLE)->get_type().get_appearance().get_rail_elevation();
	rotation = p.rotation;
	tilt = p.tilt;

	if(bogies.size()>=2)
	{
		OrientedPoint front_point = placement.get_bogie_point(bogies.front().type->index);
		bogies.front().direction = front_point.rotation-p.rotation;

		OrientedPoint back_point = placement.get_bogie_point(bogies.back().type->index);
		bogies.back().direction = back_point.rotation-p.rotation;
	}

	if(!prev)
		check_sensor(placement.get_position(VehiclePlacement::FRONT_AXLE), front_sensor, sign<0);
	if(!next)
		check_sensor(placement.get_position(VehiclePlacement::BACK_AXLE), back_sensor, sign>0);

	signal_moved.emit();
}

void Vehicle::update_position_from(const Vehicle &veh)
{
	int sign = (&veh==prev ? -1 : 1);

	float tdist = (type.get_length()+veh.type.get_length())/2;
	float margin = layout.get_catalogue().get_scale();

	float dist = distance(veh.position, position);
	if(!is_placed() || dist<tdist-margin || dist>tdist+margin)
	{
		if(sign<0)
			placement.place_after(veh.placement);
		else
			placement.place_before(veh.placement);
		update_position(0);

		dist = distance(veh.position, position);
	}

	float d = sign*(tdist-dist);
	placement.advance(d);
	update_position(d<0 ? -1 : 1);
	turn_axles(d);
}

void Vehicle::propagate_position()
{
	if(prev)
		propagate_forward();
	if(next)
		propagate_backward();
}

void Vehicle::propagate_forward()
{
	prev->update_position_from(*this);

	if(prev->prev)
		prev->propagate_forward();
}

void Vehicle::propagate_backward()
{
	next->update_position_from(*this);

	if(next->next)
		next->propagate_backward();
}

void Vehicle::check_sensor(const TrackOffsetIter &t, unsigned &sensor, bool release)
{
	unsigned s = t->get_sensor_address();
	if(s!=sensor)
	{
		unsigned old = sensor;
		sensor = s;
		if(release)
			layout.get_driver().set_sensor(old, false);
		else
			layout.get_driver().set_sensor(sensor, true);
	}
}

void Vehicle::turn_axles(float d)
{
	for(vector<Axle>::iterator i=axles.begin(); i!=axles.end(); ++i)
		i->angle += Angle::from_radians(d*2/i->type->wheel_dia);

	update_rods();
}

void Vehicle::update_rods()
{
	if(rods.empty())
		return;

	for(unsigned n=0; n<10; ++n)
	{
		float max_d = 0;
		for(vector<Rod>::iterator i=rods.begin(); i!=rods.end(); ++i)
		{
			const vector<VehicleType::RodConstraint> &constraints = i->type->constraints;
			for(vector<VehicleType::RodConstraint>::const_iterator j=constraints.begin(); j!=constraints.end(); ++j)
			{
				float d = resolve_rod_constraint(*i, *j);
				max_d = max(d, max_d);
			}
		}

		if(max_d<0.0001)
			break;
	}
}

float Vehicle::resolve_rod_constraint(Rod &rod, const VehicleType::RodConstraint &cns)
{
	Vector target;
	if(cns.target==VehicleType::RodConstraint::BODY)
		target = cns.target_position;
	else if(cns.target==VehicleType::RodConstraint::BOGIE)
		;  // TODO currently rods must lie in the xz plane of the body
	else if(cns.target==VehicleType::RodConstraint::AXLE)
	{
		const Axle &axle = get_axle(cns.target_index);
		target = Vector(axle.type->position, 0, axle.type->wheel_dia/2);
		target += Transform::rotation(axle.angle, Vector(0, 1, 0)).transform(cns.target_position);
	}
	else if(cns.target==VehicleType::RodConstraint::ROD)
	{
		const Rod &trod = get_rod(cns.target_index);
		target = trod.position;
		target += Transform::rotation(trod.angle, Vector(0, -1, 0)).transform(cns.target_position);
	}

	Vector old_position = rod.position;
	if(cns.type==VehicleType::RodConstraint::MOVE)
		rod.position = target-Transform::rotation(rod.angle, Vector(0, -1, 0)).transform(cns.local_position);
	else if(cns.type==VehicleType::RodConstraint::SLIDE)
	{
		Vector d = rod.position-target;
		rod.position = target+cns.axis*dot(d, cns.axis);
		rod.angle = Angle::zero();
	}
	else if(cns.type==VehicleType::RodConstraint::ROTATE)
	{
		Angle old_angle = rod.angle;
		Vector d = target-rod.position;
		rod.angle = Geometry::atan2<float>(d.z, d.x);
		if(cns.local_position.x || cns.local_position.z)
			rod.angle -= Geometry::atan2<float>(cns.local_position.z, cns.local_position.x);
		return abs(rod.angle-old_angle).radians()*cns.local_position.norm();
	}

	return distance(old_position, rod.position);
}

unsigned Vehicle::get_n_link_slots() const
{
	return 2;
}

Vehicle *Vehicle::get_link(unsigned i) const
{
	if(i>=2)
		throw out_of_range("Vehicle::get_link");

	return (i==0 ? prev : next);
}

int Vehicle::get_link_slot(const Object &other) const
{
	if(&other==prev)
		return 0;
	else if(&other==next)
		return 1;
	else
		return -1;
}

bool Vehicle::collide_ray(const Ray &ray, float *d) const
{
	if(is_placed())
		return Object::collide_ray(ray, d);
	else
		return false;
}


Vehicle::Axle::Axle(const VehicleType::Axle &t):
	type(&t)
{ }


Vehicle::Bogie::Bogie(const VehicleType::Bogie &t):
	type(&t),
	axles(t.axles.size())
{ }


Vehicle::Rod::Rod(const VehicleType::Rod &t):
	type(&t),
	position(t.initial_position)
{ }

} // namespace R2C2
